from typing import Any
from typing import Generator
from typing import Sequence
from typing import Union
from typing import cast

from config import MISSING
from gyver.attrs import define
from lazyfields import lazyfield
from pydantic import BaseModel
from pydantic import v1
from pydantic.fields import FieldInfo

from gyver.config.adapter.interface import FieldResolverStrategy


@define
class FieldWrapper:
    name: str
    field_info: FieldInfo


@define
class PydanticResolverStrategy(FieldResolverStrategy[FieldWrapper]):
    field: FieldWrapper

    @lazyfield
    def field_info(self) -> FieldInfo:
        return self.field.field_info

    def cast(self) -> type:
        return cast(type, self.field_info.annotation)

    def names(self) -> Sequence[str]:
        return (self.field.name, self.field_info.alias or self.field.name)

    def init_name(self) -> str:
        return self.field.name

    def default(self) -> Union[Any, type[MISSING]]:
        if self.field_info.default not in (None, Ellipsis):
            return self.field_info.default
        return (
            MISSING
            if self.field_info.default_factory is None
            else self.field_info.default_factory
        )

    @classmethod
    def iterfield(
        cls,
        config_class: type[Union[BaseModel, v1.BaseModel]],
    ) -> Generator[FieldWrapper, Any, Any]:
        if issubclass(config_class, v1.BaseModel):
            yield from cls._as_v1_iterfield(config_class)
        else:
            yield from map(
                lambda item: FieldWrapper(*item), config_class.model_fields.items()
            )

    @classmethod
    def _as_v1_iterfield(
        cls, config_class: type[v1.BaseModel]
    ) -> Generator[FieldWrapper, Any, Any]:
        for field in config_class.__fields__.values():
            field_info = FieldInfo(
                annotation=field.outer_type_,
                default=field.default,
                default_factory=field.default_factory,
                alias=field.alias,
            )
            fw = FieldWrapper(field.name, field_info)
            yield fw



from collections import defaultdict
from typing import Mapping, Optional
from urllib.parse import parse_qs, quote

from typing_extensions import Self

from gyver.attrs import mutable
from gyver.url.encode import Encodable


@mutable(eq=False)
class Query(Encodable):
    """Represents a query string and provides methods to manipulate and encode it.

    Attributes:
        params (defaultdict[str, list[str]]): The dictionary containing query parameters.
    """

    params: defaultdict[str, list[str]]

    def __init__(self, querystr: str) -> None:
        """Initializes the Query object with the provided query string.

        Args:
            querystr (str): The input query string to parse.
        """
        self.params = defaultdict(list)
        self.params |= parse_qs(querystr, keep_blank_values=True)

    def encode(self) -> str:
        """Encodes the query object into a string representation.

        Returns:
            str: The encoded query string.
        """
        return "&".join(
            "=".join((quote(key, safe=""), quote(item or "", safe="")))
            for key, value in self.params.items()
            for item in value
        )

    def omit_empty_equal(self) -> str:
        """Encodes the query object into a string representation, omitting the equal sign if a value is empty.

        Returns:
            str: The encoded query string.
        """
        return "&".join(
            "=".join(map(quote, (key, item))) if item else quote(key)
            for key, value in self.params.items()
            for item in value
        )

    def add(self, args: Optional[Mapping[str, str]] = None, /, **params: str) -> Self:
        """Adds query parameters to the Query object.

        Args:
            args (Optional[Mapping[str, str]], optional): Additional query parameters as a mapping object.
            **params (str): Additional query parameters as keyword arguments.

        Returns:
            Query: The modified Query object.
        """
        query = {**(args or {}), **params}
        for key, value in query.items():
            self.params[key].append(value)
        return self

    def set(self, args: Optional[Mapping[str, str]] = None, /, **params: str) -> Self:
        """Sets the query parameters, replacing any existing parameters.

        Args:
            args (Optional[Mapping[str, str]], optional): New query parameters as a mapping object.
            **params (str): New query parameters as keyword arguments.

        Returns:
            Query: The modified Query object.
        """
        query = {**(args or {}), **params}
        self.params.clear()
        self.add(query)
        return self

    def __setitem__(self, key: str, value: str) -> None:
        """Sets a query parameter using the square bracket syntax.

        Args:
            key (str): The parameter key.
            value (str): The parameter value.
        """
        self.add({key: value})

    def __getitem__(self, key: str) -> list[str]:
        """Retrieves the values of a query parameter using the square bracket syntax.

        Args:
            key (str): The parameter key.

        Returns:
            list[str]: The list of parameter values.
        """
        return self.params[key]

    def remove(self, *keys: str) -> Self:
        """Removes query parameters from the Query object.

        Args:
            keys (str): The parameter keys to remove.

        Returns:
            Query: The modified Query object.
        """
        for key in keys:
            self.params.pop(key, None)
        return self

    def sort(self) -> Self:
        """Sorts the query parameters in lexicographic order based on the parameter names.

        Returns:
            Query: The modified Query object.
        """
        self.params = defaultdict(list, sorted(self.params.items()))
        return self

    def copy(self) -> "Query":
        query = object.__new__(type(self))
        query.params = self.params.copy()
        return query

    def first(self) -> dict[str, str]:
        """Return a dictionary containing the first value for each query parameter.

        Returns:
            dict[str, str]: A dictionary where keys are query parameter names
                and values are the first values associated with those parameters.
        """
        return {key: value[0] for key, value in self.params.items()}

    def index_or_first(self, idx: int) -> dict[str, str]:
        """Return a dictionary containing the value at index 'idx' for each query parameter,
        or the first value if 'idx' exceeds the length of the parameter's value list.

        Args:
            idx (int): The index to retrieve for each query parameter.

        Returns:
            dict[str, str]: A dictionary where keys are query parameter names
                and values are the values at index 'idx' or the first values
                if 'idx' exceeds the length of the parameter's value list.
        """
        return {
            key: value[0 if len(value) <= idx else idx]
            for key, value in self.params.items()
        }

    def index_or_last(self, idx: int) -> dict[str, str]:
        """Return a dictionary containing the value at index 'idx' for each query parameter,
        or the last value if 'idx' exceeds the length of the parameter's value list.

        Args:
            idx (int): The index to retrieve for each query parameter.

        Returns:
            dict[str, str]: A dictionary where keys are query parameter names
                and values are the values at index 'idx' or the last values
                if 'idx' exceeds the length of the parameter's value list.
        """
        return {
            key: value[-1 if len(value) <= idx else idx]
            for key, value in self.params.items()
        }



from typing import Mapping, Optional
from urllib.parse import urlparse, urlunsplit

from typing_extensions import Self

from gyver.attrs import mutable
from gyver.url.encode import Encodable
from gyver.url.fragment import Fragment
from gyver.url.netloc import Netloc
from gyver.url.path import Path
from gyver.url.query import Query


@mutable(eq=False)
class URL(Encodable):
    """Class representing a URL object.

    Attributes:
        scheme (str): The URL scheme.
        netloc (Netloc): The URL netloc.
        path (Path): The URL path.
        query (Query): The URL query parameters.
        fragment (Fragment): The URL fragment.
    """

    scheme: str
    netloc: Netloc
    path: Path
    query: Query
    fragment: Fragment

    def __init__(self, val: str) -> None:
        """Initialize a URL object.

        Args:
            val (str): The URL string.
        """
        self.load(val)

    def load(self, val: str):
        """Parse the given URL string and populate the properties of the URL object.

        Args:
            val (str): The URL string.
        """
        parsed = urlparse(val)
        (
            self.scheme,
            netloc,
            path,
            _,
            query,
            fragment,
        ) = parsed
        self.query = Query(query)
        self.path = Path(path)
        self.fragment = Fragment(fragment)
        self.netloc = Netloc(netloc)

    def encode(self, append_empty_equal: bool = True) -> str:
        """Encode the URL object as a URL string.

        Args:
            append_empty_equal (bool, optional): Whether to append empty values with an equal sign in the query string.

        Returns:
            str: The encoded URL string.
        """
        resolved_query = (
            self.query.encode() if append_empty_equal else self.query.omit_empty_equal()
        )
        return urlunsplit(
            (
                self.scheme,
                self.netloc.encode(),
                self.path.encode(),
                resolved_query,
                self.fragment.encode(),
            )
        )

    def add(
        self,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_obj: Optional[Netloc] = None,
        scheme: Optional[str] = None,
    ) -> Self:
        """Add components to the URL.

        Args:
            queryasdict (Optional[Mapping[str, str]], optional): Dictionary-like object representing query parameters.
            path (Optional[str], optional): Path string to add to the URL.
            query (Optional[Mapping[str, str]], optional): Dictionary-like object representing additional query parameters.
            fragment (Optional[str], optional): Fragment string to set for the URL.
            netloc (Optional[str], optional): Netloc string to set for the URL.
            netloc_obj (Optional[Netloc], optional): Netloc object to merge with the existing netloc.
            scheme (Optional[str], optional): Scheme to set for the URL.

        Returns:
            URL: The updated URL object.
        """
        if queryasdict:
            self.query.add(queryasdict)
        if path:
            self.path.add(path)
        if query:
            self.query.add(query)
        if fragment:
            self.fragment.set(fragment)
        if netloc:
            self.netloc = self.netloc.merge(Netloc(netloc))
        if netloc_obj:
            self.netloc = self.netloc.merge(netloc_obj)
        if scheme:
            self.scheme = scheme
        return self

    def set(
        self,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_obj: Optional[Netloc] = None,
        scheme: Optional[str] = None,
    ) -> Self:
        """Set components of the URL.

        Args:
            queryasdict (Optional[Mapping[str, str]], optional): Dictionary-like object representing query parameters.
            path (Optional[str], optional): Path string to set for the URL.
            query (Optional[Mapping[str, str]], optional): Dictionary-like object representing query parameters.
            fragment (Optional[str], optional): Fragment string to set for the URL.
            netloc (Optional[str], optional): Netloc string to set for the URL.
            netloc_obj (Optional[Netloc], optional): Netloc object to set as the netloc.
            scheme (Optional[str], optional): Scheme to set for the URL.

        Returns:
            URL: The updated URL object.
        """
        if queryasdict:
            self.query.set(queryasdict)
        if path:
            self.path.set(path)
        if query:
            self.query.set(query)
        if fragment:
            self.fragment.set(fragment)
        if netloc:
            self.netloc.load(netloc)
        if netloc_obj:
            self.netloc = netloc_obj
        if scheme:
            self.scheme = scheme
        return self

    def copy(self) -> "URL":
        """Create a copy of the URL object.

        Returns:
            URL: The copied URL object.
        """
        url = object.__new__(type(self))
        url.scheme = self.scheme
        url.netloc = self.netloc.copy()
        url.path = self.path.copy()
        url.query = self.query.copy()
        url.fragment = self.fragment.copy()
        return url

    @classmethod
    def from_netloc(
        cls,
        netloc: Optional[Netloc] = None,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> Self:
        """Create a URL object from a netloc.

        Args:
            netloc (Optional[Netloc], optional): Netloc object to set for the URL.
            username (Optional[str], optional): Username for the netloc.
            password (Optional[str], optional): Password for the netloc.
            host (Optional[str], optional): Host for the netloc.
            port (Optional[int], optional): Port for the netloc.

        Returns:
            URL: The URL object created from the netloc.
        """
        url = cls("")
        newnetloc = Netloc.from_args(host or "", username, password, port)
        if netloc:
            newnetloc = netloc.merge(newnetloc)
        url.add(netloc_obj=newnetloc)
        return url

    @classmethod
    def from_args(
        cls,
        queryasdict: Optional[Mapping[str, str]] = None,
        /,
        path: Optional[str] = None,
        query: Optional[Mapping[str, str]] = None,
        fragment: Optional[str] = None,
        netloc: Optional[str] = None,
        netloc_obj: Optional[Netloc] = None,
        scheme: Optional[str] = None,
    ):
        url = URL("")
        url.set(
            queryasdict,
            path,
            query,
            fragment,
            netloc,
            netloc_obj,
            scheme,
        )
        return url



import functools
import warnings
from collections.abc import Callable
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from gyver.misc import functions, json, merge_dicts, strings, timezone
from gyver.misc.exc import scream

T = TypeVar("T")
P = ParamSpec("P")
ExcT = TypeVar("ExcT", bound=Exception)

to_snake = strings.to_snake
to_pascal = strings.to_pascal
to_camel = strings.to_camel
upper_camel = strings.upper_camel
cache = functions.cache


def panic(exc: type[ExcT], message: str, *args) -> ExcT:
    """
    Create an instance of the specified exception class with a modified error message.

    This function creates an exception instance by calling the constructor of the
    specified exception class with the modified error message and any additional
    arguments provided.

    Args:
        exc (type[T]): The exception class to instantiate.
        message (str): The error message for the exception.
        *args (Any): Additional arguments to pass to the exception constructor.

    Returns:
        T: An instance of the specified exception class.

    Examples:
        >>> class CustomError(Exception):
        ...     pass
        ...
        >>> error = panic(CustomError, "Something went wrong", "additional_info")
        >>> isinstance(error, CustomError)
        True
        >>> str(error)
        'Something went wrong! additional_info'
    """
    return scream(exc, message, *args)


def deprecated(func: Callable[P, T]) -> Callable[P, T]:
    """
    Mark a function as deprecated and issue a warning when it's used.

    Args:
        func (Callable[P, T]): The function to be marked as deprecated.

    Returns:
        Callable[P, T]: A wrapped version of the function that issues a warning on use.
    """

    @functools.wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        if not hasattr(func, "__warn_deprecated__"):
            warnings.warn(
                f"{func.__qualname__} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            func.__warn_deprecated__ = True
        return func(*args, **kwargs)

    return inner


if TYPE_CHECKING:
    import typing_extensions as te

    deprecated = te.deprecated  # type: ignore


class DeprecatedClass:
    """
    A class to mark as deprecated.

    This class issues a deprecation warning when instantiated.
    """

    __warn_deprecated__ = False

    def __init__(self) -> None:
        if not type(self).__warn_deprecated__:
            warnings.warn(
                f"{'.'.join((type(self).__module__, type(self).__name__))} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            type(self).__warn_deprecated__ = True


__all__ = [
    "json",
    "merge_dicts",
    "timezone",
    "strings",
    "functions",
    "panic",
    "to_snake",
    "to_pascal",
    "to_camel",
    "upper_camel",
    "cache",
    "deprecated",
    "DeprecatedClass",
]



from typing import Any, Generator, Sequence, Union

from attr._make import Factory
from attrs import NOTHING, Attribute, fields
from config import MISSING
from config.exceptions import InvalidCast

from gyver.attrs import define
from gyver.config.adapter.interface import FieldResolverStrategy
from gyver.utils import panic


@define
class AttrsResolverStrategy(FieldResolverStrategy[Attribute]):
    field: Attribute

    def cast(self) -> type:
        if field_type := self.field.type:
            return field_type
        raise panic(InvalidCast, f"Field {self.field.name} has no valid type")

    def names(self) -> Sequence[str]:
        return tuple(
            item for item in (self.field.name, self.field.alias) if item is not None
        )

    def init_name(self) -> str:
        return self.field.alias or self.field.name

    def default(self) -> Union[Any, type[MISSING]]:
        default = self.field.default
        if default is None or default in (Ellipsis, NOTHING):
            return MISSING
        return default.factory() if isinstance(default, Factory) else default

    @staticmethod
    def iterfield(config_class: type) -> Generator[Attribute, Any, Any]:
        yield from fields(config_class)



from .bound import AsyncBoundContext
from .bound import BoundContext
from .core import AsyncAtomicContext
from .core import AtomicContext
from .resolver import atomic

__all__ = [
    "atomic",
    "BoundContext",
    "AsyncBoundContext",
    "AtomicContext",
    "AsyncAtomicContext",
]



from typing import TYPE_CHECKING
from typing import Any
from typing import Generator
from typing import Protocol
from typing import Sequence
from typing import TypeVar
from typing import Union

from config import MISSING

if TYPE_CHECKING:
    from dataclasses import Field

    from attrs import Attribute
    from gyver.attrs.field import Field as GField

    from .pydantic import FieldWrapper

T_co = TypeVar("T_co", "FieldWrapper", "Attribute", "Field", "GField", covariant=True)


class FieldResolverStrategy(Protocol[T_co]):
    """
    A protocol for resolving the properties of a field, regardless of
    its implementation.

    The `FieldResolver` protocol is defined to be generic over `T`,
    which is restricted to be one of three possible types:
    - `ModelField` from the `pydantic` library
    - `Attribute` from the `attrs` library
    - `Field` from the `dataclasses` library

    These types indicate the different ways that a field can be defined in Python.
    """

    def __init__(self, field: T_co) -> None:
        """
        Initialize a new `FieldResolver` instance.

        :param field: An instance of one of the three types `T` can be.
        """
        ...

    def names(self) -> Sequence[str]:
        """
        Get the names by which the field can be accessed.

        :return: A sequence of strings, representing the names.
        """
        ...

    def init_name(self) -> str:
        """
        Get the name by which the field is expected in init.

        :return: A sequence of strings, representing the names.
        """
        ...

    def cast(self) -> type:
        """
        Get the type of the field.

        :return: The type of the field.
        """
        ...

    def default(self) -> Union[Any, type[MISSING]]:
        """
        Get the default value of the field.

        :return: The default value of the field,
        or the special value `MISSING` (from the `config` module)
        if no default is defined.
        """
        ...

    @staticmethod
    def iterfield(config_class: type) -> Generator[T_co, Any, Any]:
        ...



from gyver.exc import InvalidPath


def python_filename(filename: str, private: bool = False, dunder: bool = False):
    if private and dunder:
        raise InvalidPath("cannot have both private and dunder file")
    elif private:
        filename = f"_{filename}"
    elif dunder:
        filename = f"__{filename}__"
    return f"{filename}.py"
